import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.Scanner;
import java.util.regex.Pattern;

/**
 * SessionOrganizer reads in a similarity matrix of papers, and organizes them
 * into sessions and tracks.
 * 
 * @author janara
 *
 */
public class SessionOrganizer {
    private double[][] distanceMatrix;
    
    private int parallelTracks = 0;
    private int papersInSession = 0;
    private int sessionsInTrack = 0;
    
    private Conference conference;
    
    private double processingTimeInMinutes = 0;
    private double tradeoffCoefficient = 1.0; // the tradeoff coefficient
    
    public SessionOrganizer(String filename) {
        try {
            readInInputFile(filename);
        } catch (FileNotFoundException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        
        conference = new Conference(parallelTracks, sessionsInTrack, papersInSession);
    }
    
    /**
     * Read in the number of parallel tracks, papers in session, sessions
     * in a track, and the similarity matrix from the specified filename.
     * @param filename is the name of the file containing the matrix.
     * @return the similarity matrix.
     * @throws FileNotFoundException
     */
    public void readInInputFile(String filename) 
            throws FileNotFoundException {
        Scanner in = new Scanner(new File(filename));
        ArrayList<String> lines = new ArrayList<String>();
        while (in.hasNextLine()) {
            String line = in.nextLine().trim();
            if (!line.equals("")) {
                lines.add(line);
            }
        }
        
        if (6 > lines.size()) {
            System.err.println("Not enough information given, check format of input file");
            System.exit(1);
        }
        processingTimeInMinutes = Double.parseDouble(lines.get(0));
        papersInSession = Integer.parseInt(lines.get(1));
        parallelTracks = Integer.parseInt(lines.get(2));
        sessionsInTrack = Integer.parseInt(lines.get(3));
        tradeoffCoefficient = Double.parseDouble(lines.get(4));

        int n = lines.size()-5;
        double[][] tempDistanceMatrix = new double[n][n];
        for (int i = 0; i < n; i++) {
            String tempLine = lines.get(i+5);
            String[] elements = tempLine.split(" ");
            if (elements.length != n) {
                System.err.println(n+" "+elements.length);
                System.err.println("The similarity matrix does not have the correct format.");
                System.exit(1);
            }
            for (int j = 0; j < n; j++) {
                tempDistanceMatrix[i][j] = Double.parseDouble(elements[j].trim());
            }
        }
        distanceMatrix = tempDistanceMatrix;
        
        int numberOfPapers = distanceMatrix.length;
        int slots = parallelTracks*papersInSession*sessionsInTrack;
        if (slots != numberOfPapers) {
            System.err.println("More papers than slots available! slots:"+slots+" num papers:"+numberOfPapers);
            System.exit(1);
        }
    }
    
    /**
     * Organize the papers according to some algorithm.
     */
    public void organizePapers() {
        int paperCounter = 0;
        for (int i = 0; i < conference.getSessionsInTrack(); i++) {
            for (int j = 0; j < conference.getParallelTracks(); j++) {
                for (int k = 0; k < conference.getPapersInSession(); k++) {
                    conference.setPaper(j, i, k, paperCounter);
                    paperCounter++;
                }
            }
        }
    }
    
    /**
     * Score the organization.
     * @return the score.
     */
    public double scoreOrganization() {

        // Sum of pairwise similarities per session.
        double score1 = 0.0;
        for (int i = 0; i < conference.getParallelTracks(); i++) {
            Track tmpTrack = conference.getTrack(i);
            for (int j = 0; j < tmpTrack.getNumberOfSessions(); j++) {
                Session tmpSession = tmpTrack.getSession(j);
                for (int k = 0; k < tmpSession.getNumberOfPapers(); k++) {
                    int index1 = tmpSession.getPaper(k);
                    for (int l = k+1; l < tmpSession.getNumberOfPapers(); l++) {
                        int index2 = tmpSession.getPaper(l);
                        //System.out.println("sim:"+index1+" "+index2+" "+(1-distanceMatrix[index1][index2]));
                        score1 += 1-distanceMatrix[index1][index2];
                    }
                }
            }
        }
        
        // Sum of distances for competing papers.
        double score2 = 0.0;
        for (int i = 0; i < conference.getParallelTracks(); i++) {
            Track tmpTrack1 = conference.getTrack(i);
            for (int j = 0; j < tmpTrack1.getNumberOfSessions(); j++) {
                Session tmpSession1 = tmpTrack1.getSession(j);
                for (int k = 0; k < tmpSession1.getNumberOfPapers(); k++) {
                    int index1 = tmpSession1.getPaper(k);

                    // Get competing papers.
                    for (int l = i+1; l < conference.getParallelTracks(); l++) {
                        Track tmpTrack2 = conference.getTrack(l);
                        Session tmpSession2 = tmpTrack2.getSession(j);
                        for (int m = 0; m < tmpSession2.getNumberOfPapers(); m++) {
                            int index2 = tmpSession2.getPaper(m);
                            //System.out.println("dist:"+index1+" "+index2+" "+(distanceMatrix[index1][index2]));
                            score2 += distanceMatrix[index1][index2];
                        }
                    }
                }
            }
        }
        
        double score = score1+tradeoffCoefficient*score2;

        return score;
    }
    
    /**
     * Get the distance matrix.
     * @return the distance matrix.
     */
    public double[][] getDistanceMatrix() {
        return distanceMatrix;
    }
    
    public String toString() {
        return conference.toString();
    }
    
    public void readInOrganization(String filename) {
        Pattern p = Pattern.compile("\\|");
        Scanner in = null;
        try {
            in = new Scanner(new File(filename));
        } catch (FileNotFoundException e) {
            System.err.println("File "+filename+" does not exist.");
            System.exit(1);
        }
        int sessionIndex = 0;
        while (in.hasNextLine()) {
            String line = in.nextLine().trim();
            String[] sessions = p.split(line);
            if (!line.equals("")) {
                for (int trackIndex = 0; trackIndex < sessions.length; trackIndex++) {
                    String[] papers = sessions[trackIndex].trim().split(" ");
                    for (int paperIndex = 0; paperIndex < papers.length; paperIndex++) {
                        int paperId = Integer.parseInt(papers[paperIndex]);
                        conference.setPaper(trackIndex, sessionIndex, paperIndex, paperId);
                    }
                }
            }
            sessionIndex++;
        }
    }
    
    public static void main(String args[]) {
        
        // Parse the input.
        if (args.length < 1) {
            System.err.println("SessionOrganizer <input_filename>");
            System.exit(1);
        }
        String inputfilename = args[0];
        String organizationfilename = "";
        if (args.length == 2) {
            organizationfilename = args[1];
        }
        // Initialize the conference organizer.
        SessionOrganizer organizer = new SessionOrganizer(inputfilename);
        
        if (organizationfilename.equals("")) {
            // Organize the papers into tracks based on similarity.
            organizer.organizePapers();
        } else {
            // Read in an organization from a file
            organizer.readInOrganization(organizationfilename);
        }
        System.out.println("organization:");
        System.out.println(organizer);
        
        // Score the organization against the gold standard.
        double score = organizer.scoreOrganization();
        System.out.println("score:"+score);
    }
}
